import sys

import django

from django_migrant.management.commands.migrant import Command

django.setup()

# Mimic args when the command is called as a django-admin command.
# Required because commands assume positional values in sys.argv.
argv = ["django-admin", "migrant"] + sys.argv[1:]
Command().run_from_argv(argv)



import json
import unittest
from unittest import mock

from django.db.migrations.graph import Node

from django_migrant.management.commands import migrant


class TestStageTwo(unittest.TestCase):

    @mock.patch("django_migrant.management.commands.migrant.subprocess", mock.MagicMock())
    @mock.patch("django_migrant.management.commands.migrant.Path", mock.MagicMock())
    @mock.patch("django_migrant.management.commands.migrant.call_command")
    @mock.patch("django_migrant.management.commands.migrant.MigrationLoader")
    def test_migrate_to_zero(self, mock_loader, mock_call_command):
        # The give node 0001_initial has no parent, so we migrate to the
        # "zero" migration of that app.
        n1 = Node(("polls", "0001_initial"))
        mock_loader.return_value.graph.node_map = {n1.key: n1}
        file_contents = json.dumps(
            [
                ["polls", "0001_initial"],
            ]
        )
        with mock.patch(
            "builtins.open", mock.mock_open(read_data=file_contents)
        ) as mock_open:
            migrant.stage_two()
        handle = mock_open()
        handle.read.assert_called_once()

        mock_call_command.assert_called_once()
        self.assertTrue(len(mock_call_command.call_args_list), 1)
        self.assertEqual(mock_call_command.call_args.args, ("migrate", "polls", "zero"))

    @mock.patch("django_migrant.management.commands.migrant.subprocess", mock.MagicMock())
    @mock.patch("django_migrant.management.commands.migrant.Path", mock.MagicMock())
    @mock.patch("django_migrant.management.commands.migrant.call_command")
    @mock.patch("django_migrant.management.commands.migrant.MigrationLoader")
    def test_migrate_to_parent(self, mock_loader, mock_call_command):
        # To reverse 0002 we migrate to 0001.

        n1 = Node(("polls", "0001_initial"))
        n2 = Node(("polls", "0002_alter_question_question_text"))
        n2.add_parent(n1)
        mock_loader.return_value.graph.node_map = {n1.key: n1, n2.key: n2}
        file_contents = json.dumps([["polls", "0002_alter_question_question_text"]])
        with mock.patch(
            "builtins.open", mock.mock_open(read_data=file_contents)
        ) as mock_open:
            migrant.stage_two()
        handle = mock_open()
        handle.read.assert_called_once()

        mock_call_command.assert_called_once()
        self.assertTrue(len(mock_call_command.call_args_list), 1)
        self.assertEqual(
            mock_call_command.call_args.args, ("migrate", "polls", "0001_initial")
        )

    @mock.patch("django_migrant.management.commands.migrant.subprocess", mock.MagicMock())
    @mock.patch("django_migrant.management.commands.migrant.Path", mock.MagicMock())
    @mock.patch("django_migrant.management.commands.migrant.call_command")
    @mock.patch("django_migrant.management.commands.migrant.MigrationLoader")
    def test_migrate_to_single_ancestor(self, mock_loader, mock_call_command):
        # The provided file gives a list of migrations in the same app. Make sure
        # we only migrate to one of them.

        n1 = Node(("polls", "0001_initial"))
        n2 = Node(("polls", "0002_alter_question_question_text"))
        n2.add_parent(n1)
        n3 = Node(("polls", "0003_alter_title"))
        n3.add_parent(n2)
        n4 = Node(("polls", "0004_alter_description"))
        n4.add_parent(n3)
        mock_loader.return_value.graph.node_map = {
            n1.key: n1,
            n2.key: n2,
            n3.key: n3,
            n4.key: n4,
        }
        file_contents = json.dumps(
            [
                ["polls", "0002_alter_question_question_text"],
                ["polls", "0003_alter_title"],
                ["polls", "0004_alter_description"],
            ]
        )
        with mock.patch(
            "builtins.open", mock.mock_open(read_data=file_contents)
        ) as mock_open:
            migrant.stage_two()
        handle = mock_open()
        handle.read.assert_called_once()

        mock_call_command.assert_called_once()
        self.assertTrue(len(mock_call_command.call_args_list), 1)
        self.assertEqual(
            mock_call_command.call_args.args, ("migrate", "polls", "0001_initial")
        )



import django
from django.test import SimpleTestCase

# Settings configuration has to be outside of test cases
# to allow test discovery to work.
# May need to wrap in exception catcher, and/or put in test class.


class DjangoSetupTestCase(SimpleTestCase):

    def setUp(self):

        django.setup()

        return super().setUp()



import json
import os
import subprocess
import sys
from importlib import resources
from pathlib import Path

from django.core.management import call_command
from django.core.management.base import BaseCommand, CommandError
from django.db import DEFAULT_DB_ALIAS, connections
from django.db.migrations.loader import MigrationLoader

MIGRANT_FILENAME = Path(".") / ".migrant"


def stage_one():
    connection = connections[DEFAULT_DB_ALIAS]
    loader = MigrationLoader(connection)
    targets = set(loader.applied_migrations) - set(loader.disk_migrations)

    targets_as_json = json.dumps(list(targets))
    with open(MIGRANT_FILENAME, "w+") as fh:
        fh.write(targets_as_json)

    env_with_stage_two = os.environ.copy()
    env_with_stage_two["DJANGO_MIGRANT_STAGE"] = "TWO"
    # NB: We use raw subprocess because dulwich (used to interface with git repos)
    # doesn't support the relative branch "-".
    subprocess.run(["git", "checkout", "-", "--quiet"], env=env_with_stage_two)


def stage_two():
    connection = connections[DEFAULT_DB_ALIAS]
    loader = MigrationLoader(connection)
    with open(MIGRANT_FILENAME) as fh:
        node_names = [tuple(n) for n in json.loads(fh.read())]

    nodes = [loader.graph.node_map[n] for n in node_names]

    targets = set()
    for n in nodes:
        if not n.parents:
            targets.add((n.key[0], "zero"))
        else:
            for parent in n.parents:
                if parent not in nodes:
                    targets.add(parent.key)
    for t in list(targets):
        call_command("migrate", t[0], t[1])

    env_with_stage_three = os.environ.copy()
    env_with_stage_three["DJANGO_MIGRANT_STAGE"] = "THREE"
    subprocess.run(["git", "checkout", "-", "--quiet"], env=env_with_stage_three)


def stage_three():
    call_command("migrate")


class Command(BaseCommand):
    def add_arguments(self, parser):
        subparsers = parser.add_subparsers(
            title="sub-commands",
            required=True,
        )

        install_parser = subparsers.add_parser(
            "install",
            help="Installs the command to run in a 'post-checkout' git hook.",
        )
        install_parser.set_defaults(method=self.install)
        install_parser.add_argument("dest")
        install_parser.add_argument("-i", "--interpreter", default=sys.executable)

        migrate_parser = subparsers.add_parser(
            "migrate",
            help="Migrates database seemlessly from one git branch to another.",
        )

        migrate_parser.set_defaults(method=self.migrate)

    def handle(self, *args, method, **options):
        method(*args, **options)

    def install(self, *args, **options):
        path = Path(options["dest"])
        git_path = path / ".git"

        # Validate the given destination.
        if not git_path.is_dir():
            raise CommandError(f"'{path}' does not appear to contain a git repo.")

        dest_git_hooks_path = git_path / "hooks"
        if not dest_git_hooks_path:
            raise CommandError(f"'{path}' does not contain a 'hooks' directory.")

        context = {"interpreter": options["interpreter"]}
        self.create_hook(dest_git_hooks_path, "post-checkout", context)
        self.create_hook(dest_git_hooks_path, "pre-rebase")

    def create_hook(self, path: Path, name: str, context: dict = None):
        if context is None:
            context = {}
        hook_filename = path / name
        if hook_filename.is_file():
            ok_to_append = input(
                f"'{name}' hook already exists. Append to file [y/N]? "
            )
            if ok_to_append.upper() != "Y":
                raise CommandError(f"'{path}' already contains a post-checkout hook.")
            exists = True
        else:
            exists = False

        template_filename = resources.files("django_migrant") / "hook_templates" / name

        with open(template_filename, "r") as fh:
            content = fh.read()

        # A poor mans templating system.
        for key, value in context.items():
            content = content.replace("{{ %s }}" % key, value)

        if not exists:
            header_filename = (
                resources.files("django_migrant") / "hook_templates" / "header"
            )
            with open(header_filename, "r") as fh:
                header = fh.read()

            header = header.replace("{{ name }}", str(name))
        else:
            header = ""

        content = header + content
        with open(hook_filename, "a") as fh:
            fh.write(content)

        self.stdout.write(f"{name} hook created: {hook_filename}")

    def migrate(self, *args, **options):
        DJANGO_MIGRANT_STAGE = os.environ.get("DJANGO_MIGRANT_STAGE")
        if not DJANGO_MIGRANT_STAGE:
            stage_one()
        elif DJANGO_MIGRANT_STAGE == "TWO":
            stage_two()
        elif DJANGO_MIGRANT_STAGE == "THREE":
            stage_three()



from unittest import mock

from tests.testcases import DjangoSetupTestCase


class MainTests(DjangoSetupTestCase):
    # Test invocation of command directly, not via django-admin.
    @mock.patch("sys.argv", ["1", "2", "3"])
    @mock.patch("django_migrant.management.commands.migrant.Command")
    def test_simple(self, mock_command):
        # Trick python into invoking __main__ code.
        from django_migrant import __main__  # noqa: F401

        mock_command.assert_called_once()
        mock_command.return_value.run_from_argv.assert_called_with
        mock_command.return_value.run_from_argv.assert_called_with(
            ["django-admin", "migrant", "2", "3"]
        )



import os
from importlib import resources
from io import StringIO
from pathlib import Path
from tempfile import TemporaryDirectory
from unittest import mock

from django.core.management import call_command
from django.core.management.base import CommandError

from tests.testcases import DjangoSetupTestCase


def get_mock_path(is_dir=False, is_file=False, is_true=False):
    """Creates a mock Path object that has scoped parameters."""

    class MockPath(mock.Mock):
        # Use Mock and not MagicMock so that magic methods can be provided.
        def __init__(self, *args, **kwargs):
            kwargs["spec_set"] = Path
            super().__init__(*args, **kwargs)

        def __truediv__(self, other):
            return MockPath()

        def __bool__(self):
            return is_true

        def is_dir(self):
            return is_dir

        def is_file(self):
            return is_file

    return MockPath()


class CommandTests(DjangoSetupTestCase):

    def call_command(self, *args, **kwargs):
        out = StringIO()
        err = StringIO()
        call_command(
            "migrant",
            *args,
            stdout=out,
            stderr=err,
            **kwargs,
        )
        return out.getvalue(), err.getvalue()

    def test_install(self):

        templates_dir = resources.files("django_migrant") / "hook_templates"

        with open(templates_dir / "header", "r") as fh:
            header = fh.read()

        with TemporaryDirectory() as temp_dir_name:
            hooks_path = Path(temp_dir_name) / ".git" / "hooks"
            hooks_path.mkdir(parents=True)
            out, err = self.call_command("install", temp_dir_name)

            with open(hooks_path / "post-checkout") as fh:
                contents = fh.read()
                self.assertTrue(contents.startswith(header[:9]))
                # A phrase we definitely expect to see in the hook.
                self.assertTrue("django_migrant migrate" in contents)

            with open(hooks_path / "pre-rebase") as fh:
                contents = fh.read()
                self.assertTrue(contents.startswith(header[:9]))
                self.assertTrue('echo "REBASE" > .migrant' in contents)

        output = out.split("\n")
        # Remove the runt line.
        output = [x for x in output if x]
        self.assertEqual(len(output), 2)
        # The order of output doesn't really matter.
        self.assertTrue(output[0].startswith("post-checkout hook created: "))
        self.assertTrue(output[1].startswith("pre-rebase hook created: "))
        self.assertEqual(err, "")

    @mock.patch(
        "django_migrant.management.commands.migrant.Path", get_mock_path(is_dir=False)
    )
    def test_install_not_git_dir(self):
        with self.assertRaises(CommandError) as context:
            self.call_command("install", "/a/destination/")
        msg = str(context.exception)
        self.assertTrue("does not appear to contain a git repo" in msg)

    @mock.patch(
        "django_migrant.management.commands.migrant.Path",
        get_mock_path(is_dir=True, is_true=False),
    )
    def test_install_no_githooks_path(self):
        with self.assertRaises(CommandError) as context:
            self.call_command("install", "/a/destination/")
        msg = str(context.exception)
        self.assertTrue("does not contain a 'hooks' directory" in msg)

    @mock.patch(
        "django_migrant.management.commands.migrant.Path",
        get_mock_path(is_dir=True, is_true=True, is_file=True),
    )
    @mock.patch("builtins.input")
    def test_install_file_exists_dont_append(self, mock_input):
        # User presses 'N' when asked to append.
        mock_input.return_value = "N"
        with self.assertRaises(CommandError) as context:
            self.call_command("install", "/a/destination/")
        msg = str(context.exception)
        self.assertTrue("already contains a post-checkout hook" in msg)

    @mock.patch("builtins.input")
    def test_install_file_exists_do_append(self, mock_input):
        # Check behaviour when appending to an existing hook.

        # User presses 'y' when asked to append.
        mock_input.return_value = "y"

        # Prepare an existing post-checkout hook
        # Perhaps this is a better than the get_mock_path approach?
        with TemporaryDirectory() as temp_dir_name:
            hooks_path = Path(temp_dir_name) / ".git" / "hooks"
            hooks_path.mkdir(parents=True)
            post_checkout_file = hooks_path / "post-checkout"

            with open(post_checkout_file, "w") as fh:
                fh.write("Existing file contents\n")

            self.call_command("install", temp_dir_name)

            with open(post_checkout_file, "r") as fh:
                contents = fh.read()
                # Original content is still there.
                self.assertTrue("Existing file contents" in contents)

                # And content has been appended.
                self.assertTrue("START django_migrant" in contents)

    @mock.patch("django_migrant.management.commands.migrant.stage_one")
    def test_migrate_stage_one(self, mock_stage_one):
        out, err = self.call_command("migrate")
        self.assertEqual(err, "")

        mock_stage_one.assert_called_once()

    @mock.patch.dict(os.environ, {"DJANGO_MIGRANT_STAGE": "TWO"})
    @mock.patch("django_migrant.management.commands.migrant.stage_two")
    def test_migrate_stage_two(self, mock_stage_two):
        out, err = self.call_command("migrate")
        self.assertEqual(err, "")

        mock_stage_two.assert_called_once()

    @mock.patch.dict(os.environ, {"DJANGO_MIGRANT_STAGE": "THREE"})
    @mock.patch("django_migrant.management.commands.migrant.stage_three")
    def test_migrate_stage_three(self, mock_stage_three):
        out, err = self.call_command("migrate")
        self.assertEqual(err, "")

        mock_stage_three.assert_called_once()
        mock_stage_three.assert_called_once()



from unittest import mock

from django_migrant.management.commands import migrant
from tests.testcases import DjangoSetupTestCase


class TestStageOne(DjangoSetupTestCase):

    @mock.patch("django_migrant.management.commands.migrant.subprocess", mock.MagicMock())
    @mock.patch("django_migrant.management.commands.migrant.Path", mock.MagicMock())
    @mock.patch("django_migrant.management.commands.migrant.MigrationLoader")
    def test_file_written(self, mock_loader):
        mock_loader.return_value.applied_migrations = [
            ("polls", "0002_alter_question_question_text"),
            ("polls", "0001_initial"),
        ]

        mock_loader.return_value.disk_migrations = [
            ("polls", "0001_initial"),
        ]

        with mock.patch("builtins.open", mock.mock_open()) as mock_open:
            migrant.stage_one()
        handle = mock_open()
        handle.write.assert_called_once_with(
            '[["polls", "0002_alter_question_question_text"]]'
        )

    @mock.patch("django_migrant.management.commands.migrant.Path", mock.MagicMock())
    @mock.patch(
        "django_migrant.management.commands.migrant.MigrationLoader", mock.MagicMock()
    )
    @mock.patch("django_migrant.management.commands.migrant.subprocess")
    def test_subprocess_called(self, mock_subprocess):

        with mock.patch("builtins.open", mock.mock_open()):
            migrant.stage_one()

        # Subprocess was called.
        mock_subprocess.run.assert_called_once()
        self.assertEqual(len(mock_subprocess.run.call_args), 2)

        # And was called with git checkout.
        call_args = mock_subprocess.run.call_args
        first_arg = call_args[0]
        self.assertEqual(first_arg[0], ["git", "checkout", "-", "--quiet"])

        # And env var is set to stage two.
        second_arg = call_args[1]
        self.assertTrue("env" in second_arg)
        self.assertTrue("DJANGO_MIGRANT_STAGE" in second_arg["env"])
        self.assertEqual(second_arg["env"]["DJANGO_MIGRANT_STAGE"], "TWO")



import unittest
from unittest import mock

from django_migrant.management.commands import migrant


class TestStageThree(unittest.TestCase):

    @mock.patch("django_migrant.management.commands.migrant.call_command")
    def test_command_called(self, mock_call_command):

        migrant.stage_three()

        mock_call_command.assert_called_once()
        self.assertTrue(len(mock_call_command.call_args_list), 1)
        self.assertEqual(mock_call_command.call_args.args, ("migrate",))
